/***************************************************************************
 * Copyright 2013 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.service.impl;

import java.io.IOException;
import java.util.Collection;
import java.util.List;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;

import kieker.analysis.AnalysisController.STATE;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.webgui.common.ClassAndMethodContainer;
import kieker.webgui.common.exception.AnalysisInitializationException;
import kieker.webgui.common.exception.AnalysisStateException;
import kieker.webgui.common.exception.DisplayNotFoundException;
import kieker.webgui.common.exception.NewerProjectException;
import kieker.webgui.common.exception.ProjectAlreadyExistingException;
import kieker.webgui.common.exception.ProjectNotExistingException;
import kieker.webgui.domain.ComponentListContainer;
import kieker.webgui.persistence.IProjectDAO;
import kieker.webgui.service.IProjectService;
import kieker.webgui.service.impl.util.ACManager;

import org.primefaces.model.UploadedFile;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

/**
 * This is an implemented facade for the project management. For technical reasons it is a singleton class.
 * 
 * @author Nils Christian Ehmke
 */
// TODO Remove the lock objects for removed projects as well (take a look at #591).
@Service
public class ProjectServiceImpl implements IProjectService {

	private final ConcurrentHashMap<String, Object> fileSystemLocks = new ConcurrentHashMap<String, Object>();
	private final ConcurrentHashMap<String, Object> analysesLocks = new ConcurrentHashMap<String, Object>();
	@Autowired
	private ACManager acManager;
	@Autowired
	private IProjectDAO projectDAO;

	/**
	 * Default constructor. <b>Do not use this constructor. This bean is Spring managed.</b>
	 */
	public ProjectServiceImpl() {
		// No code necessary.
	}

	@Override
	public void updateAllAnalyses() {
		for (final Entry<String, Object> locks : this.analysesLocks.entrySet()) {
			synchronized (locks.getValue()) {
				this.acManager.updateDisplays(locks.getKey());
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#addProject(java.lang.String, java.lang.String)
	 */
	@Override
	public void addProject(final String projectName, final String username) throws ProjectAlreadyExistingException, IOException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			this.projectDAO.addProject(projectName, username);
		}
	}

	@Override
	public void delProject(final String projectName) throws ProjectNotExistingException, IOException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			this.projectDAO.removeProject(projectName);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#copyProject(java.lang.String, java.lang.String)
	 */
	@Override
	public void copyProject(final String originalProjectName, final String newProjectName) throws ProjectNotExistingException, ProjectAlreadyExistingException,
			IOException {
		// Get the file system locks for both projects
		final Object srcProjectLock = this.getLock(originalProjectName, this.fileSystemLocks);
		final Object dstProjectLock = this.getLock(newProjectName, this.fileSystemLocks);

		// We cannot risk just using two synchronized-blocks as such a thing could block the application. We therefore have to sort the locks by name in order to
		// make sure that two methods would use the same order to access (if project A has to be renamed to B and B has to be renamed to A, we would always
		// synchronize on A before synchronizing on B).
		final Object lockFst;
		final Object lockSnd;
		if (originalProjectName.compareTo(newProjectName) < 0) {
			lockFst = srcProjectLock;
			lockSnd = dstProjectLock;
		} else {
			lockFst = dstProjectLock;
			lockSnd = srcProjectLock;
		}

		synchronized (lockFst) {
			synchronized (lockSnd) {
				this.projectDAO.copyProject(originalProjectName, newProjectName);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#openProject(java.lang.String)
	 */
	@Override
	public MIProject openProject(final String projectName) throws ProjectNotExistingException, IOException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			return this.projectDAO.openProject(projectName);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#openProject(java.lang.String, kieker.webgui.common.ClassAndMethodContainer)
	 */
	@Override
	public Object openProject(final String projectName, final ClassAndMethodContainer classAndMethodContainer) throws ProjectNotExistingException, IOException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			return this.projectDAO.openProject(projectName, classAndMethodContainer);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#saveProject(java.lang.String, kieker.analysis.model.analysisMetaModel.MIProject, long, boolean, String, String)
	 */
	@Override
	public void saveProject(final String projectName, final MIProject project, final long timeStamp, final boolean overwriteNewerProject,
			final String username, final String analysisLayout, final String cockpitLayout) throws ProjectNotExistingException, IOException, NewerProjectException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			this.projectDAO.saveProject(projectName, project, timeStamp, overwriteNewerProject, username, analysisLayout, cockpitLayout);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#getCurrTimeStamp(java.lang.String)
	 */
	@Override
	public long getCurrTimeStamp(final String projectName) throws ProjectNotExistingException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			return this.projectDAO.getCurrTimeStamp(projectName);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#uploadLibrary(org.primefaces.model.UploadedFile, java.lang.String)
	 */
	@Override
	public void uploadLibrary(final UploadedFile file, final String projectName) throws ProjectNotExistingException, IOException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			this.projectDAO.uploadLibrary(file, projectName);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#getClassLoader(java.lang.String, java.lang.Object)
	 */
	@Override
	public ClassLoader getClassLoader(final String projectName, final Object requester) throws ProjectNotExistingException, IOException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			return this.projectDAO.getClassLoader(projectName, requester); // NOPMD (ClassLoader)
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#getAvailableComponents(java.lang.String)
	 */
	@Override
	public ComponentListContainer getAvailableComponents(final String projectName) {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			return this.projectDAO.getAvailableComponents(projectName);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#listAllLibraries(java.lang.String)
	 */
	@Override
	public List<String> listAllLibraries(final String projectName) throws ProjectNotExistingException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			return this.projectDAO.listAllLibraries(projectName);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#listAllProjects()
	 */
	@Override
	public Collection<String> listAllProjects() {
		return this.projectDAO.listAllProjects();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#initializeAnalysis(java.lang.String, java.lang.ClassLoader)
	 */
	@Override
	public void initializeAnalysis(final String projectName, final ClassLoader classLoader) throws ProjectNotExistingException, AnalysisStateException,
			AnalysisInitializationException {
		// We have to lock both - the project and the analysis, as a file has to be loaded
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);
		final Object analysisLock = this.getLock(projectName, this.analysesLocks);

		synchronized (projectLock) {
			synchronized (analysisLock) {
				this.acManager.initializeAnalysis(projectName, classLoader);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#cleanAnalysis(java.lang.String)
	 */
	@Override
	public void cleanAnalysis(final String projectName) throws ProjectNotExistingException, AnalysisStateException {
		final Object analysisLock = this.getLock(projectName, this.analysesLocks);

		synchronized (analysisLock) {
			this.acManager.cleanAnalysis(projectName);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#startAnalysis(java.lang.String)
	 */
	@Override
	public void startAnalysis(final String projectName) throws ProjectNotExistingException, AnalysisStateException {
		final Object analysisLock = this.getLock(projectName, this.analysesLocks);

		synchronized (analysisLock) {
			this.acManager.startAnalysis(projectName);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#stopAnalysis(java.lang.String)
	 */
	@Override
	public void stopAnalysis(final String projectName) throws ProjectNotExistingException, AnalysisStateException {
		final Object analysisLock = this.getLock(projectName, this.analysesLocks);

		synchronized (analysisLock) {
			this.acManager.stopAnalysis(projectName);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#getDisplay(java.lang.String, java.lang.String, java.lang.String)
	 */
	@Override
	public Object getDisplay(final String projectName, final String viewName, final String displayName) throws ProjectNotExistingException,
			DisplayNotFoundException {
		final Object analysisLock = this.getLock(projectName, this.analysesLocks);

		synchronized (analysisLock) {
			return this.acManager.getDisplay(projectName, viewName, displayName);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#getCurrentState(java.lang.String)
	 */
	@Override
	public STATE getCurrentState(final String projectName) {
		final Object analysisLock = this.getLock(projectName, this.analysesLocks);

		synchronized (analysisLock) {
			return this.acManager.getCurrentState(projectName);
		}
	}

	/**
	 * This method can be used to get the lock for a given project. If the lock doesn't exist already, a new one will be created.
	 * 
	 * @param projectName
	 *            The project whose lock should be delivered.
	 * @param lockMap
	 *            The map containing the locks.
	 * @return A lock object for the project.
	 */
	private Object getLock(final String projectName, final ConcurrentHashMap<String, Object> lockMap) {
		// Create a new lock and put it into the map - if it doesn't already contain the key
		final Object newLock = new Object();
		final Object existLock = lockMap.putIfAbsent(projectName, newLock);

		// Find out which of the both locks is the correct one
		final Object lock;
		if (existLock != null) {
			lock = existLock;
		} else {
			lock = newLock;
		}

		return lock;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#getLogEntries(java.lang.String)
	 */
	@Override
	public Object[] getLogEntries(final String projectName) throws AnalysisStateException {
		final Object analysisLock = this.getLock(projectName, this.analysesLocks);

		synchronized (analysisLock) {
			return this.acManager.getLogEntries(projectName);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#deleteLibrary(java.lang.String)
	 */
	@Override
	public boolean deleteLibrary(final String projectName, final String libName) throws IOException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			return this.projectDAO.deleteLibrary(projectName, libName);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#getOwner(java.lang.String)
	 */
	@Override
	public String getOwner(final String projectName) throws ProjectNotExistingException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			return this.projectDAO.getOwner(projectName);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#getLastUser(java.lang.String)
	 */
	@Override
	public String getLastUser(final String projectName) throws ProjectNotExistingException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			return this.projectDAO.getLastUser(projectName);
		}
	}

	@Override
	public void importProject(final String projectName, final String username, final UploadedFile file) throws ProjectAlreadyExistingException, IOException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			this.projectDAO.importProject(projectName, username, file);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#getAnalysisLayout(java.lang.String)
	 */
	@Override
	public String getAnalysisLayout(final String projectName) {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			return this.projectDAO.getAnalysisLayout(projectName);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.IProjectService#getCockpitLayout(java.lang.String)
	 */
	@Override
	public String getCockpitLayout(final String projectName) {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			return this.projectDAO.getCockpitLayout(projectName);
		}
	}
}
