/***************************************************************************
 * Copyright 2013 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.service.impl.util;

import java.io.File;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;

import kieker.analysis.AnalysisController;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.webgui.common.ClassAndMethodContainer;
import kieker.webgui.common.exception.AnalysisInitializationException;
import kieker.webgui.common.exception.AnalysisStateException;
import kieker.webgui.common.exception.ProjectLoadException;

import net.vidageek.mirror.dsl.Mirror;
import net.vidageek.mirror.exception.MirrorException;

/**
 * An analysis within the web gui.
 * 
 * @author Nils Christian Ehmke
 */
public class Analysis {

	private static final Log LOG = LogFactory.getLog(Analysis.class);

	private static final long MAX_THREAD_WAIT_TIME_MS = 1000;

	private final ClassAndMethodContainer classAndMethodContainer;
	private final Object analysisController;
	private final Object analysisControllerThread;

	private final Map<String, Map<String, Object>> displayObjects = new HashMap<String, Map<String, Object>>();
	private final Map<String, Map<String, Method>> displayMethods = new HashMap<String, Map<String, Method>>();

	/**
	 * Creates a new instance of this class using the given parameters.
	 * 
	 * @param classLoader
	 *            The class loader to be used to initialize the analysis.
	 * @param projectFile
	 *            The file to be loaded.
	 * @throws AnalysisStateException
	 *             If something went wrong during the loading of the analysis.
	 * @throws AnalysisInitializationException
	 *             If an error occurred during the instantiation of the analysis.
	 */
	public Analysis(final ClassLoader classLoader, final File projectFile) throws AnalysisStateException, AnalysisInitializationException {
		try {
			this.classAndMethodContainer = new ClassAndMethodContainer(classLoader);

			final Method createMethod = this.classAndMethodContainer.getAnalysisControllerCreateAnalysisController();
			final Object modelProject = ClassAndMethodContainer.invokeClassMethod(this.classAndMethodContainer.getAnalysisControllerLoadFromFile(), null,
					projectFile.getAbsoluteFile());
			final Object controllerAndMapping = ClassAndMethodContainer.invokeClassMethod(createMethod, null, modelProject, classLoader);
			this.analysisController = ClassAndMethodContainer.invokeMethod(this.classAndMethodContainer.getAnalysisControllerWithMappingGetController(),
					controllerAndMapping, null);
			this.analysisControllerThread = ClassAndMethodContainer.invokeConstructor(this.classAndMethodContainer.getAnalysisControllerThreadConstructor(),
					this.analysisController);
			// this.updateDisplaysThread = new UpdateDisplaysThread(); // ClassAndMethodContainer.invokeMethod(
			// this.classAndMethodContainer.getAnalysisControllerWithMappingGetMapping(), controllerAndMapping, null));

			if ((this.analysisController == null) || (this.analysisControllerThread == null)) {
				throw new AnalysisInitializationException("An error occured while instantiating the analysis.");
			}
		} catch (final ProjectLoadException ex) {
			throw new AnalysisInitializationException("An error occured while instantiating the analysis.", ex);
		} catch (final NullPointerException ex) {
			throw new AnalysisInitializationException("An error occured while instantiating the analysis.", ex);
		}
	}

	/**
	 * Starts the analysis.
	 * 
	 * @throws AnalysisStateException
	 *             If the analysis is in the wrong state to be started.
	 */
	public void start() throws AnalysisStateException {
		try {
			new Mirror().on(this.analysisControllerThread).invoke().method("start").withoutArgs();
		} catch (final MirrorException ex) {
			Analysis.LOG.error("An error occured during a reflection method call.", ex);
		}
	}

	/**
	 * Stops the analysis.
	 */
	public void stop() {
		try {
			new Mirror().on(this.analysisControllerThread).invoke().method("terminate").withoutArgs();
			new Mirror().on(this.analysisControllerThread).invoke().method("join").withArgs(Analysis.MAX_THREAD_WAIT_TIME_MS);
		} catch (final MirrorException ex) {
			Analysis.LOG.error("An error occured during a reflection method call.", ex);
		}
	}

	/**
	 * Delivers the current state of the analysis.
	 * 
	 * @return The state.
	 */
	public Enum<?> getCurrentState() {
		try {
			return (Enum<?>) new Mirror().on(this.analysisController).invoke().method("getState").withoutArgs();
		} catch (final MirrorException ex) {
			Analysis.LOG.error("An error occured during a reflection method call.", ex);
			return null;
		}
	}

	/**
	 * This method delivers the display object of the (currently running) analysis. Technically it is an instance of {@code AbstractDisplay}, but in fact the project
	 * specific class loader has been used.
	 * 
	 * @param viewName
	 *            The name of the view.
	 * @param displayName
	 *            The name of the display.
	 * @return A display object for the given parameters.
	 */
	public Object getDisplay(final String viewName, final String displayName) {
		// return this.updateDisplaysThread.getDisplay(viewName, displayName);
		return null;
	}

	/**
	 * Delivers the log entries for the current analysis.
	 * 
	 * @return An array containing the log entries (if available).
	 */
	public Object[] getLogEntries() {
		return (Object[]) ClassAndMethodContainer.invokeClassMethod(Analysis.this.classAndMethodContainer.getLogImplWebguiLoggingClassGetEntriesMethod(), null,
				AnalysisController.class.getName());
	}

	public void updateDisplays() {

	}
}
