/***************************************************************************
 * Copyright 2013 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.web.beans.view;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import javax.faces.application.FacesMessage;
import javax.faces.context.FacesContext;
import javax.faces.event.ValueChangeEvent;

import kieker.analysis.model.analysisMetaModel.MIAnalysisComponent;
import kieker.analysis.model.analysisMetaModel.MIAnalysisMetaModelFactory;
import kieker.analysis.model.analysisMetaModel.MIDependency;
import kieker.analysis.model.analysisMetaModel.MIFilter;
import kieker.analysis.model.analysisMetaModel.MIInputPort;
import kieker.analysis.model.analysisMetaModel.MIOutputPort;
import kieker.analysis.model.analysisMetaModel.MIPlugin;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.analysis.model.analysisMetaModel.MIReader;
import kieker.analysis.model.analysisMetaModel.MIRepository;
import kieker.analysis.model.analysisMetaModel.MIRepositoryConnector;
import kieker.analysis.model.analysisMetaModel.impl.MAnalysisMetaModelFactory;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.webgui.common.exception.LibraryLoadException;
import kieker.webgui.common.exception.NewerProjectException;
import kieker.webgui.common.exception.ProjectLoadException;
import kieker.webgui.common.exception.ProjectNotExistingException;
import kieker.webgui.domain.ComponentListContainer;
import kieker.webgui.domain.pluginDecorators.AbstractAnalysisComponentDecorator;
import kieker.webgui.domain.pluginDecorators.FilterDecorator;
import kieker.webgui.domain.pluginDecorators.ReaderDecorator;
import kieker.webgui.domain.pluginDecorators.RepositoryDecorator;
import kieker.webgui.service.IProjectService;
import kieker.webgui.web.beans.application.GlobalPropertiesBean;
import kieker.webgui.web.beans.application.ProjectsBean;
import kieker.webgui.web.beans.session.UserBean;

import org.primefaces.context.RequestContext;
import org.primefaces.event.FileUploadEvent;
import org.primefaces.model.UploadedFile;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

/**
 * The {@link CurrentAnalysisEditorBean} contains the necessary data behind an instance of the analysis editor. It provides various methods to manipulate the current
 * project, as the analysis editor is the most important part of the whole application. The connection to the graph within the editor is done via another bean (the
 * {@link CurrentAnalysisEditorGraphBean}).<br>
 * The class is a Spring managed bean with view scope to make sure that one user (even in one session) can open multiple projects at a time without causing any
 * problems.
 * 
 * @author Nils Christian Ehmke
 */
@Component
@Scope("view")
public class CurrentAnalysisEditorBean {

	private static final Log LOG = LogFactory.getLog(CurrentAnalysisEditorBean.class);
	private static final MIAnalysisMetaModelFactory FACTORY = MAnalysisMetaModelFactory.eINSTANCE;

	private ComponentListContainer availableComponents;
	private MIAnalysisComponent selectedComponent;
	private boolean unsavedModifications;
	private MIProject project;
	private String projectName;
	private long timeStamp;

	@Autowired
	private CurrentAnalysisEditorGraphBean currentAnalysisEditorGraphBean;
	@Autowired
	private GlobalPropertiesBean globalPropertiesBean;
	@Autowired
	private IProjectService projectService;
	@Autowired
	private ProjectsBean projectsBean;
	@Autowired
	private UserBean userBean;
	private String currentLayout;

	/**
	 * Creates a new instance of this class. <b>Do not call this constructor manually. It will only be accessed by Spring.</b>
	 */
	public CurrentAnalysisEditorBean() {
		this.availableComponents = new ComponentListContainer(Collections.<ReaderDecorator>emptyList(), Collections.<FilterDecorator>emptyList(),
				Collections.<RepositoryDecorator>emptyList());
	}

	/**
	 * This method initializes the bean by using the current project name to load the project. <b>Do not call this method manually. It will only be accessed by
	 * Spring.</b>
	 */
	public void initialize() {
		try {
			// Make sure that the initialization will only be done for the init request. During all other requests, the method call have to be ignored.
			if (!FacesContext.getCurrentInstance().isPostback()) {
				// Load the project itself
				this.project = this.projectsBean.openProject(this.projectName);
				// Remember the current time! This is important for the later comparison of the time stamps.
				this.resetTimeStamp();
				// Add the libraries within the lib-folder to the current model
				this.initializeModelLibraries();
				// Load the available readers, filters and repositories
				this.reloadAvailableComponents();

				this.unsavedModifications = false;
			}
		} catch (final ProjectLoadException ex) {
			CurrentAnalysisEditorBean.LOG.error("An error occured while loading the project.", ex);
			// FacesContext.getCurrentInstance().getApplication().getNavigationHandler().handleNavigation(FacesContext.getCurrentInstance(), null,
			// "projectOverview");
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_ERROR, this.globalPropertiesBean.getMsgProjectLoadingException());
		} catch (final NullPointerException ex) {
			// This exception can occur, when a property has not been initialized
			CurrentAnalysisEditorBean.LOG.error("An error occured while loading the project.", ex);
			// FacesContext.getCurrentInstance().getApplication().getNavigationHandler().handleNavigation(FacesContext.getCurrentInstance(), null,
			// "projectOverview");
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_ERROR, this.globalPropertiesBean.getMsgProjectLoadingException());
		}
	}

	/**
	 * This method takes all libraries from the lib-folder and adds them to the in-memory-model.
	 * 
	 * @throws ProjectLoadException
	 *             If something went wrong during the loading of the libraries.
	 */
	private void initializeModelLibraries() throws ProjectLoadException {
		try {
			final List<String> libs = this.projectService.listAllLibraries(this.projectName);
			// Add them, but remove all existing dependencies so far to avoid double entries. This also makes sure that the model - after it has been opened -
			// points just to valid dependencies (and to all of them).
			this.project.getDependencies().clear();
			for (final String lib : libs) {
				final MIDependency dep = CurrentAnalysisEditorBean.FACTORY.createDependency();
				dep.setFilePath(lib);
				this.project.getDependencies().add(dep);
			}
		} catch (final ProjectNotExistingException ex) {
			throw new ProjectLoadException("The project does not exist.", ex);
		}
	}

	/**
	 * This method is the handler for the file upload. It tries to upload the given file and informs the user via the growl-component.
	 * 
	 * @param event
	 *            The upload event.
	 */
	public void handleFileUpload(final FileUploadEvent event) {
		// Get the file from the event
		final UploadedFile file = event.getFile();

		try {
			// Use the file system manager to upload the new file
			final MIDependency lib;
			this.projectService.uploadLibrary(file, this.projectName);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_INFO, this.globalPropertiesBean.getMsgLibraryUploaded());
			// As it seem to have worked, we can add the library to our model.
			lib = CurrentAnalysisEditorBean.FACTORY.createDependency();
			lib.setFilePath(file.getFileName());
			this.project.getDependencies().add(lib);
			// Update our class loader and the available plugins & repositories

			// We have to reinitialize the tool palette completely! This is necessary as some of the already existing classes could need the newly loaded classes.
			this.reloadAvailableComponents();

			this.setModificationsFlag();
		} catch (final IOException ex) {
			CurrentAnalysisEditorBean.LOG.error("An error occured while uploading the library.", ex);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_ERROR, this.globalPropertiesBean.getMsgLibraryUploadingException());
		} catch (final ProjectLoadException ex) {
			CurrentAnalysisEditorBean.LOG.error("An error occured while uploading the library.", ex);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_ERROR, this.globalPropertiesBean.getMsgLibraryUploadingException());
		} catch (final ProjectNotExistingException ex) {
			CurrentAnalysisEditorBean.LOG.error("Project does not exist.", ex);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_ERROR, this.globalPropertiesBean.getMsgProjectNotExistingException());
		} catch (final LibraryLoadException ex) {
			CurrentAnalysisEditorBean.LOG.error("An error occured while uploading the library.", ex);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_ERROR, this.globalPropertiesBean.getMsgLibraryUploadingException());
		}
	}

	/**
	 * Removes the library with the given name and reloads the classes and the classloader. Furthermore the tool palette will be reinitalized.
	 * 
	 * @param name
	 *            The name of the library to be removed.
	 */
	public void deleteLibrary(final String name) {
		try {
			if (this.projectService.deleteLibrary(this.projectName, name)) {
				// We have to reinitialize the tool palette completely! This is necessary as some of the already existing classes could need the newly loaded
				// classes.
				this.reloadAvailableComponents();

				this.setModificationsFlag();
			}
		} catch (final IOException ex) {
			CurrentAnalysisEditorBean.LOG.error("An error occured while removing the library.", ex);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while removing the library.");
		}
	}

	/**
	 * This method loads the list of available readers, filters and repositories, using the current libraries within the model.
	 */
	private void reloadAvailableComponents() {
		this.availableComponents = this.projectService.getAvailableComponents(this.projectName);
	}

	public MIProject getProject() {
		return this.project;
	}

	public void setProjectName(final String newName) {
		// Remember the given parameters
		this.projectName = newName;
	}

	public String getProjectName() {
		return this.projectName;
	}

	public ComponentListContainer getAvailableComponents() {
		return this.availableComponents;
	}

	/**
	 * This method sets the time stamp to the current system time.
	 */
	public void resetTimeStamp() {
		this.timeStamp = System.currentTimeMillis();
	}

	public long getTimeStamp() {
		return this.timeStamp;
	}

	public MIAnalysisComponent getSelectedPlugin() {
		return this.selectedComponent;
	}

	/**
	 * This method delivers the available libraries of this project as a pair of strings. The first element is always the kieker-library.
	 * 
	 * @return The available libraries.
	 */
	public List<String> getLibraries() {
		try {
			final List<String> result = this.projectService.listAllLibraries(this.projectName);
			result.add(0, "Kieker");

			return result;
		} catch (final ProjectNotExistingException ex) {
			CurrentAnalysisEditorBean.LOG.error("The project does not exist.", ex);
		} catch (final NullPointerException ex) {
			// This exception occurs when a property has not been initialized correctly.
			CurrentAnalysisEditorBean.LOG.error("A null pointer exception occured.", ex);
		}
		return new ArrayList<String>();
	}

	/**
	 * This method tries to save the current project and informs the user about success or fail. There should be two parameters within the request parameter map
	 * (layoutString and overwriteNewerProject) as this method is called via javascript.
	 */
	public void saveProject() {
		// Get the parameters
		final Map<String, String> paramMap = FacesContext.getCurrentInstance().getExternalContext().getRequestParameterMap();
		this.currentLayout = paramMap.get("layoutString");
		final boolean overwriteNewerProject = Boolean.parseBoolean(paramMap.get("overwriteNewerProject"));

		try {
			this.projectService.saveProject(this.projectName, this.project, this.timeStamp, overwriteNewerProject, this.userBean.getUsername(), this.currentLayout,
					null);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_INFO, this.globalPropertiesBean.getMsgProjectSaved());
			// Update the time stamp!
			this.resetTimeStamp();

			this.clearModificationsFlag();
		} catch (final IOException ex) {
			CurrentAnalysisEditorBean.LOG.error("An error occured while saving the project.", ex);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_ERROR, this.globalPropertiesBean.getMsgProjectSavingException());
		} catch (final NewerProjectException ex) {
			CurrentAnalysisEditorBean.LOG.info("The project has been modified externally in the meanwhile.", ex);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_WARN, this.globalPropertiesBean.getMsgProjectModified());
			// Give the user the possibility to force-save the project
			RequestContext.getCurrentInstance().execute("forceSaveDlg.show()");
		} catch (final ProjectNotExistingException ex) {
			CurrentAnalysisEditorBean.LOG.error("The project does not exist.", ex);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_ERROR, this.globalPropertiesBean.getMsgProjectNotExistingException());
		}
	}

	/**
	 * This method adds a new repository to the current model, using the given instance of {@code RepositoryContainer} for it.
	 * 
	 * @param container
	 *            The container which delivers the copy of the repository.
	 */
	public void addRepository(final RepositoryDecorator container) {
		// Create a new instance for the model
		final MIRepository repository = container.newCopy(CurrentAnalysisEditorBean.FACTORY);

		// Add it to the project - and to the graph
		this.project.getRepositories().add(repository);
		this.currentAnalysisEditorGraphBean.addRepository(repository);

		this.setModificationsFlag();
	}

	/**
	 * This method adds a new plugin to the current model, using the given instance of {@code PluginContainer} for it.
	 * 
	 * @param container
	 *            The container which delivers the copy of the plugin.
	 */
	public void addPlugin(final FilterDecorator container) {
		// Create a new instance for the model
		final MIPlugin plugin = container.newCopy(CurrentAnalysisEditorBean.FACTORY);

		// Add it to the project - and to the graph
		this.project.getPlugins().add(plugin);
		if (plugin instanceof MIReader) {
			this.currentAnalysisEditorGraphBean.addReader((MIReader) plugin);
		} else {
			this.currentAnalysisEditorGraphBean.addFilter((MIFilter) plugin);
		}

		this.setModificationsFlag();
	}

	/**
	 * This method sets the property {@link CurrentAnalysisEditorBean#unsavedModifications} to false and refreshes the necessary components within the analysis
	 * editor to make this visible.
	 */
	private void clearModificationsFlag() {
		this.unsavedModifications = false;
		RequestContext.getCurrentInstance().update("menuForm");
	}

	/**
	 * This method sets the property {@link CurrentAnalysisEditorBean#unsavedModifications} to true and refreshes the necessary components within the analysis editor
	 * to make this visible.
	 */
	private void setModificationsFlag() {
		this.unsavedModifications = true;
		RequestContext.getCurrentInstance().update("menuForm");
	}

	/**
	 * Getter for the property {@link CurrentAnalysisEditorBean#unsavedModifications}.
	 * 
	 * @return The current value of the property.
	 */
	public boolean isUnsavedModification() {
		return this.unsavedModifications;
	}

	/**
	 * This method delivers the properties of the currently selected plugin, but it adds also the name- and the class-properties as a string to the list. The first
	 * element is always the classname, the second one is always the pluginname.
	 * 
	 * @return A list with all properties of the plugin plus the name- and class-properties. If no plugin is selected, the list of the selected repository is
	 *         delivered.
	 */
	public List<Object> getAdvancedPluginProperties() {
		final List<Object> result = new ArrayList<Object>();

		// Add the properties as strings
		result.add("Name");
		result.add("ClassName");

		// Get the original properties of the plugin
		if (this.selectedComponent instanceof MIPlugin) {
			result.addAll(((MIPlugin) this.selectedComponent).getProperties());
		} else {
			result.addAll(((MIRepository) this.selectedComponent).getProperties());
		}

		return result;
	}

	/**
	 * Delivers a human readable description of the given property.
	 * 
	 * @param component
	 *            The parent of the property.
	 * @param property
	 *            The property name.
	 * @return A human readable description and a substitution if there is no description.
	 */
	public String getDescription(final MIAnalysisComponent component, final String property) {
		AbstractAnalysisComponentDecorator<? extends MIAnalysisComponent> container = null;

		// Find the container which contains the component
		if (component instanceof MIReader) {
			final String className = ((MIReader) component).getClassname();
			for (final ReaderDecorator reader : this.availableComponents.getReaders()) {
				if (reader.getClassname().equals(className)) {
					container = reader;
					break;
				}
			}
		} else if (component instanceof MIFilter) {
			final String className = ((MIFilter) component).getClassname();
			for (final FilterDecorator filter : this.availableComponents.getFilters()) {
				if (filter.getClassname().equals(className)) {
					container = filter;
					break;
				}
			}
		} else {
			final String className = ((MIRepository) component).getClassname();
			for (final RepositoryDecorator repository : this.availableComponents.getRepositories()) {
				if (repository.getClassname().equals(className)) {
					container = repository;
					break;
				}
			}
		}

		// Extract the description
		if (container != null) {
			return container.getPropertyDescription(property);
		}

		// Nothing found
		return "No description available.";
	}

	/**
	 * This method should be called if the grid color has been modified.
	 * 
	 * @param event
	 *            The change event.
	 */
	public void gridColorListener(final ValueChangeEvent event) {
		this.currentAnalysisEditorGraphBean.setGridColor((String) event.getNewValue());
	}

	/**
	 * This method should be called if the grid size has been modified.
	 * 
	 * @param event
	 *            The change event.
	 */
	public void gridSizeListener(final ValueChangeEvent event) {
		this.currentAnalysisEditorGraphBean.setGridSize((Integer) event.getNewValue());
	}

	/**
	 * This method initializes the modified jit-graph by delivering the necessary javascript commands to the client. It prints all current existing plugins,
	 * repositories and their connections. All further components will be added mostly by other methods.
	 */
	public void initializeGraph() {
		// Make sure that the currentAnalysisEditorGraphBean knows "this" as well.
		this.currentAnalysisEditorGraphBean.setCurrentAnalysisEditorBean(this);

		// Initialize the graph
		this.currentAnalysisEditorGraphBean.declareGraph();

		// Initialize the reader, filter and repositories
		for (final MIPlugin plugin : this.project.getPlugins()) {
			if (plugin instanceof MIReader) {
				this.currentAnalysisEditorGraphBean.addReader((MIReader) plugin);
			} else {
				this.currentAnalysisEditorGraphBean.addFilter((MIFilter) plugin);
			}
		}

		for (final MIRepository repository : this.project.getRepositories()) {
			this.currentAnalysisEditorGraphBean.addRepository(repository);
		}

		// Now initialize the connections between filters...
		for (final MIPlugin plugin : this.project.getPlugins()) {
			for (final MIOutputPort oPort : plugin.getOutputPorts()) {
				for (final MIInputPort iPort : oPort.getSubscribers()) {
					this.currentAnalysisEditorGraphBean.addConnection(plugin, iPort.getParent(), oPort, iPort);
				}
			}
		}

		// ...and between filters and repositories
		for (final MIPlugin plugin : this.project.getPlugins()) {
			for (final MIRepositoryConnector rPort : plugin.getRepositories()) {
				// It is possible that the connected repository is null, if it hasn't been set yet. Check this.
				if (rPort.getRepository() != null) {
					this.currentAnalysisEditorGraphBean.addConnection(plugin, rPort.getRepository(), rPort);
				}
			}
		}

		this.currentAnalysisEditorGraphBean.initListeners();
		this.currentAnalysisEditorGraphBean.addEdgeConstraints();

		// Now we have to set the default grid size and color of the user
		this.currentAnalysisEditorGraphBean.setGridColor(this.userBean.getGridColor());
		this.currentAnalysisEditorGraphBean.setGridSize(this.userBean.getGridSize());

		// Perform either an initial auto layout or use the saved layout - if it exists
		final String layout = this.projectService.getAnalysisLayout(this.projectName);

		if (layout != null) {
			this.currentAnalysisEditorGraphBean.loadLayout(layout);
		} else {
			this.currentAnalysisEditorGraphBean.startAutoLayout();
		}

		this.currentAnalysisEditorGraphBean.checkReadOnlyForGuests();
	}

	/**
	 * This method should be called if a node (plugin, repository) has been selected.
	 * 
	 * @param node
	 *            The new node to be selected.
	 */
	public void nodeSelected(final MIAnalysisComponent node) {
		this.selectedComponent = node;
	}

	/**
	 * This method should be called if a node (plugin, repository) has been removed.
	 * 
	 * @param node
	 *            The new node to be removed.
	 */
	public void nodeRemoved(final MIAnalysisComponent node) {
		// Remove the component from the project
		if (node instanceof MIPlugin) {
			this.project.getPlugins().remove(node);

			// Remove the corresponding connections
			final List<MIInputPort> toBeRemoved = new ArrayList<MIInputPort>();
			for (final MIPlugin plugin : this.project.getPlugins()) {
				for (final MIOutputPort oPort : plugin.getOutputPorts()) {
					toBeRemoved.clear();
					for (final MIInputPort iPort : oPort.getSubscribers()) {
						if (iPort.getParent() == node) {
							toBeRemoved.add(iPort);
						}
					}
					oPort.getSubscribers().removeAll(toBeRemoved);
				}
			}
		} else {
			this.project.getRepositories().remove(node);

			// Remove the corresponding connections
			for (final MIPlugin plugin : this.project.getPlugins()) {
				for (final MIRepositoryConnector repoConn : plugin.getRepositories()) {
					if (repoConn.getRepository() == node) {
						repoConn.setRepository(null);
					}
				}
			}
		}

		// Deselect the currently selected node if it is the one which has just been removed
		if (this.selectedComponent == node) {
			this.selectedComponent = null; // NOPMD
		}

		this.setModificationsFlag();
	}

	/**
	 * This method should be delivered if an edge between two plugins has been created.
	 * 
	 * @param sourcePort
	 *            The source port.
	 * @param targetPort
	 *            The target port.
	 */
	public void edgeCreated(final MIOutputPort sourcePort, final MIInputPort targetPort) {
		sourcePort.getSubscribers().add(targetPort);

		this.setModificationsFlag();
	}

	/**
	 * This method should be delivered if an edge between two plugins has been removed.
	 * 
	 * @param sourcePort
	 *            The source port.
	 * @param targetPort
	 *            The target port.
	 */
	public void edgeRemoved(final MIOutputPort sourcePort, final MIInputPort targetPort) {
		sourcePort.getSubscribers().remove(targetPort);

		this.setModificationsFlag();
	}

	/**
	 * This method should be delivered if an edge between a plugin and a repository has been created.
	 * 
	 * @param sourcePort
	 *            The source port.
	 * @param target
	 *            The target repository.
	 */
	public void edgeCreated(final MIRepositoryConnector sourcePort, final MIRepository target) {
		sourcePort.setRepository(target);

		this.setModificationsFlag();
	}

	/**
	 * This method should be delivered if an edge between a plugin and a repository has been removed.
	 * 
	 * @param sourcePort
	 *            The source port.
	 * @param target
	 *            The target repository.
	 */
	public void edgeRemoved(final MIRepositoryConnector sourcePort, final MIRepository target) {
		sourcePort.setRepository(null);

		this.setModificationsFlag();
	}

}
